from enum import Enum
import struct

class InverterType(Enum):
    UNDEFINED = (0, "Undefined")
    SINGLE_PHASE_HYBRID_INVERTER = (3, "Single phase hybrid inverter")
    MICRO_INVERTER = (4, "Micro inverter")
    LOW_VOLTAGE_THREE_PHASE_HYBRID_INVERTER = (5, "Low voltage three phase hybrid inverter")
    HIGH_VOLTAGE_THREE_PHASE_HYBRID_INVERTER = (6, "High voltage three phase hybrid inverter")
    HIGH_VOLTAGE_THREE_PHASE_INVERTER_6_12KW = (7, "High voltage three phase inverter 6-12kW")
    HIGH_VOLTAGE_THREE_PHASE_INVERTER_20_50WK = (262, "High voltage three phase inverter 20-50kW")
    THREE_PHASE_POWER_CONVERSION_SYSTEM = (8, "Three phase power conversion system")
    BALCONY_ENERGY_STORAGE_SYSTEM = (9, "Balcony energy storage system")

class BasicInfo:
    def __init__(self, type: InverterType, serial_number, main_version, hmi_version, protocol_version, rated_power) -> None:
        self.type = type
        self.serial_number = serial_number
        self.main_version = main_version
        self.hmi_version = hmi_version
        self.protocol_version = protocol_version
        self.rated_power = rated_power


class ModbusMapper:
    def __init__(self, register_values, start_address):
        self.register_values = register_values
        self.start_address = start_address

    def get_value(self, desired_address):
        if desired_address < self.start_address or desired_address >= self.start_address + len(self.register_values):
            return None  # Desired address is out of range

        index = desired_address - self.start_address
        return self.register_values[index]
    
    def get_uint16(self, desired_address):
        value = self.get_value(desired_address)
        return value

    def get_int16(self, desired_address):
        value = self.get_value(desired_address)
        if value is None:
            return None
        if value > 32767:
            value -= 65536
        return value        
    
    def get_uint32(self, desired_address, word_swap=True):
        high_word = self.get_value(desired_address)
        low_word = self.get_value(desired_address + 1)
        if word_swap:
            return (low_word << 16) + high_word
        return (high_word << 16) + low_word
    
    def get_string(self, desired_address):
        value = self.get_value(desired_address)
        high_byte = (value >> 8) & 0xFF
        low_byte = value & 0xFF
        return "".join([chr(high_byte), chr(low_byte)])
    
    def dump(self):
        for i, value in enumerate(self.register_values):
            print(f"Address {self.start_address + i}: {value} 0x{value:04X}")


from ..interfaces import ModbusTCP
from ..measurements import InverterMeasurements, EnergyMeasurements, BatteryMeasurements, PhaseMeasurements, ACMeasurements, PVStringMeasurements, Measurement
from ..helper import ModbusMapper
from ..inverters.base_inverter import BaseInverter


class DeyeSUN12KEU(BaseInverter):

    measurements: InverterMeasurements

    def __init__(self, adapter):
        self.adapter = adapter
        self.external_ct = False


    async def update_status(self):
        self.measurements = await self.get_measurements()


    async def get_measurements(self):
        print("Getting measurements")
        data = await self.adapter.read_registers(514, 125)
        data2 = await self.adapter.read_registers(514+125, 125)

        mapper = ModbusMapper(data+data2, 514)

        # 514 daily battery charge uint16 0.1 scale 
        daily_battery_charge_value = round(mapper.get_uint16(514) * 0.1, 1)
        daily_battery_charge = Measurement(value=daily_battery_charge_value, unit="kWh")


        # 515 daily battery discharge uint16 0.1 scale
        daily_battery_discharge_value = round(mapper.get_uint16(515) * 0.1, 1)
        daily_battery_discharge = Measurement(value=daily_battery_discharge_value, unit="kWh")

        # 516-517 total battery charge uint32 word swap 0.1 scale
        total_battery_charge_value = round(mapper.get_uint32(516) * 0.1, 1)
        total_battery_charge = Measurement(value=total_battery_charge_value, unit="kWh")

        # 518-519 total battery discharge uint32 word swap 0.1 scale
        total_battery_discharge_value = round(mapper.get_uint32(518) * 0.1, 1)
        total_battery_discharge = Measurement(value=total_battery_discharge_value, unit="kWh")

        # 520 daily grid bought uint16 0.1 scale
        daily_grid_bought_value = round(mapper.get_uint16(520) * 0.1, 1)
        daily_grid_bought = Measurement(value=daily_grid_bought_value, unit="kWh")

        # 521 daily grid sold uint16 0.1 scale
        daily_grid_sold_value = round(mapper.get_uint16(521) * 0.1, 1)
        daily_grid_sold = Measurement(value=daily_grid_sold_value, unit="kWh")

        # 522-523 total grid bought uint32 word swap 0.1 scale
        total_grid_bought_value =  round(mapper.get_uint32(522) * 0.1, 1)
        total_grid_bought = Measurement(value=total_grid_bought_value, unit="kWh")

        # 524-525 total grid sold uint32 word swap 0.1 scale
        total_grid_sold_value = round(mapper.get_uint32(524) * 0.1, 1)
        total_grid_sold = Measurement(value=total_grid_sold_value, unit="kWh")

        # 526 daily consumption uint16 0.1 scale
        daily_consumption_value = round(mapper.get_uint16(526) * 0.1, 1)
        daily_consumption = Measurement(value=daily_consumption_value, unit="kWh")

        # 527-528 total consumption uint32 word swap 0.1 scale
        total_consumption_value = round(mapper.get_uint32(527) * 0.1, 1)
        total_consumption = Measurement(value=total_consumption_value, unit="kWh")

        # 529 daily production uint16 0.1 scale
        daily_production_value = round(mapper.get_uint16(529) * 0.1, 1)
        daily_production = Measurement(value=daily_production_value, unit="kWh")

        # 534-535 total production uint32 word swap 0.1 scale
        total_production_value = round(mapper.get_uint32(534) * 0.1, 1)
        total_production = Measurement(value=total_production_value, unit="kWh")

        # 540 dc temperature uint16 0.1 scale -100 offset
        dc_temperature_value = round(mapper.get_int16(540) * 0.1 - 100, 1)
        dc_temperature = Measurement(value=dc_temperature_value, unit="°C")

        # 541 heatsink temperature uint16 0.1 scale -100 offset
        heatsink_temperature_value = round(mapper.get_int16(541) * 0.1 - 100, 1)
        heatsink_temperature = Measurement(value=heatsink_temperature_value, unit="°C")

        # 586 battery temperature uint16 0.1 scale -100 offset
        battery_temperature_value = round(mapper.get_int16(586) * 0.1 - 100, 1)
        battery_temperature = Measurement(value=battery_temperature_value, unit="°C")

        # 587 battery voltage uint16 0.01 scale
        battery_voltage_value = round(mapper.get_uint16(587) * 0.01, 2)
        battery_voltage = Measurement(value=battery_voltage_value, unit="V")

        # 588 battery soc uint16
        battery_soc_value = mapper.get_uint16(588)
        battery_soc = Measurement(value=battery_soc_value, unit="%")

        # 590 battery power uint16 
        battery_power_value = mapper.get_int16(590)
        battery_power = Measurement(value=battery_power_value, unit="W")

        # 591 battery current uint16 0.01 scale
        battery_current_value = round(mapper.get_int16(591) * 0.01, 2)
        battery_current = Measurement(value=battery_current_value, unit="A")

        # 592 battery corrected ah uint16
        battery_corrected_ah_value = mapper.get_uint16(592)
        battery_corrected_ah = Measurement(value=battery_corrected_ah_value, unit="AH")

        # 598 grid L1 voltage uint16 0.1 scale
        grid_L1_voltage_value = round(mapper.get_uint16(598) * 0.1, 1)
        grid_L1_voltage = Measurement(value=grid_L1_voltage_value, unit="V")

        # 599 grid L2 voltage uint16 0.1 scale
        grid_L2_voltage_value = round(mapper.get_uint16(599) * 0.1, 1)
        grid_L2_voltage = Measurement(value=grid_L2_voltage_value, unit="V")

        # 600 grid L3 voltage uint16 0.1 scale
        grid_L3_voltage_value = round(mapper.get_uint16(600) * 0.1, 1)
        grid_L3_voltage = Measurement(value=grid_L3_voltage_value, unit="V")

        
        # 604 grid L1 power internal uint16
        grid_L1_internal_power_value =  mapper.get_int16(604)
        grid_L1_internal_power = Measurement(value=grid_L1_internal_power_value, unit="W")
        
        # 605 grid L2 power internal uint16
        grid_L2_internal_power_value =  mapper.get_int16(605)
        grid_L2_internal_power = Measurement(value=grid_L2_internal_power_value, unit="W")

        # 606 grid L3 power internal uint16
        grid_L3_internal_power_value =  mapper.get_int16(606)
        grid_L3_internal_power = Measurement(value=grid_L3_internal_power_value, unit="W")

        # 609 grid frequency uint16 0.01 scale
        grid_frequency_value = round(mapper.get_uint16(609) * 0.01, 2)
        grid_frequency = Measurement(value=grid_frequency_value, unit="Hz")
       
        # 616 grid L1 power external uint16
        grid_L1_external_power_value =  mapper.get_int16(616)
        grid_L1_external_power = Measurement(value=grid_L1_external_power_value, unit="W")

        # 617 grid L2 power external uint16
        grid_L2_external_power_value =  mapper.get_int16(617)
        grid_L2_external_power = Measurement(value=grid_L2_external_power_value, unit="W")

        # 618 grid L3 power external uint16
        grid_L3_external_power_value =  mapper.get_int16(618)
        grid_L3_external_power = Measurement(value=grid_L3_external_power_value, unit="W")

        # 625 grid total power uint16
        grid_total_power_value = mapper.get_int16(625)
        grid_total_power = Measurement(value=grid_total_power_value, unit="W")

        # 630 grid L1 current uint16 0.01 scale
        grid_L1_current_value = round(mapper.get_int16(630) * 0.01, 2)
        grid_L1_current = Measurement(value=grid_L1_current_value, unit="A")

        # 631 grid L2 current uint16 0.01 scale
        grid_L2_current_value = round(mapper.get_int16(631) * 0.01, 2)
        grid_L2_current = Measurement(value=grid_L2_current_value, unit="A")

        # 632 grid L3 current uint16 0.01 scale
        grid_L3_current_value = round(mapper.get_int16(632) * 0.01, 2)
        grid_L3_current = Measurement(value=grid_L3_current_value, unit="A")

        # 633 inverter L1 power uint16
        inverter_L1_power_value = mapper.get_int16(633)
        inverter_L1_power = Measurement(value=inverter_L1_power_value, unit="W")

        # 634 inverter L2 power uint16
        inverter_L2_power_value = mapper.get_int16(634)
        inverter_L2_power = Measurement(value=inverter_L2_power_value, unit="W")

        # 635 inverter L3 power uint16
        inverter_L3_power_value = mapper.get_int16(635)
        inverter_L3_power = Measurement(value=inverter_L3_power_value, unit="W")

        # 644 load L1 voltage uint16 0.1 scale
        load_L1_voltage_value = round(mapper.get_uint16(644) * 0.1, 1)
        load_L1_voltage = Measurement(value=load_L1_voltage_value, unit="V")

        # 645 load L2 voltage uint16 0.1 scale
        load_L2_voltage_value = round(mapper.get_uint16(645) * 0.1, 1)
        load_L2_voltage = Measurement(value=load_L2_voltage_value, unit="V")

        # 646 load L3 voltage uint16 0.1 scale
        load_L3_voltage_value = round(mapper.get_uint16(646) * 0.1, 1)
        load_L3_voltage = Measurement(value=load_L3_voltage_value, unit="V")

        # 650 load L1 power uint16
        load_L1_power_value = mapper.get_uint16(650)
        load_L1_power = Measurement(value=load_L1_power_value, unit="W")
       
        # 651 load L2 power uint16
        load_L2_power_value = mapper.get_uint16(651)
        load_L2_power = Measurement(value=load_L2_power_value, unit="W")

        # 652 load L3 power uint16
        load_L3_power_value = mapper.get_uint16(652)
        load_L3_power = Measurement(value=load_L3_power_value, unit="W")

        # 653 load total power uint16
        load_total_power_value = mapper.get_uint16(653)
        load_total_power = Measurement(value=load_total_power_value, unit="W")

        # 655 load frequency uint16 0.01 scale
        load_frequency_value = round(mapper.get_uint16(655) * 0.01, 2)
        load_frequency = Measurement(value=load_frequency_value, unit="Hz")

        # 672 pv1 power uint16
        pv1_power_value = mapper.get_uint16(672)
        pv1_power = Measurement(value=pv1_power_value, unit="W")

        # 673 pv2 power uint16
        pv2_power_value = mapper.get_uint16(673)
        pv2_power = Measurement(value=pv2_power_value, unit="W")

        # 676 pv1 voltage uint16 0.1 scale
        pv1_voltage_value = round(mapper.get_uint16(676) * 0.1, 1)
        pv1_voltage = Measurement(value=pv1_voltage_value, unit="V")

        # 677 pv1 current uint16 0.1 scale
        pv1_current_value = round(mapper.get_uint16(677) * 0.1, 1)
        pv1_current = Measurement(value=pv1_current_value, unit="A")

        # 678 pv2 voltage uint16 0.1 scale
        pv2_voltage_value = round(mapper.get_uint16(678) * 0.1, 1)
        pv2_voltage = Measurement(value=pv2_voltage_value, unit="V")

        # 679 pv2 current uint16 0.1 scale
        pv2_current_value = round(mapper.get_uint16(679) * 0.1, 1)
        pv2_current = Measurement(value=pv2_current_value, unit="A")


        battery_measurements = BatteryMeasurements(
            temperature=battery_temperature,
            voltage=battery_voltage,
            current=battery_current,
            power=battery_power,
            soc=battery_soc
        )

        grid_measurements = ACMeasurements(
            total_power=grid_total_power,
            phases=[
                PhaseMeasurements(
                    voltage=grid_L1_voltage,
                    current=grid_L1_current,
                    power=grid_L1_external_power if self.external_ct else grid_L1_internal_power
                ),
                PhaseMeasurements(
                    voltage=grid_L2_voltage,
                    current=grid_L2_current,
                    power=grid_L2_external_power if self.external_ct else grid_L2_internal_power
                ),
                PhaseMeasurements(
                    voltage=grid_L3_voltage,
                    current=grid_L3_current,
                    power=grid_L3_external_power if self.external_ct else grid_L3_internal_power
                )
            ],
            frequency=grid_frequency
        )

        load_measurements = ACMeasurements(
            total_power=load_total_power,
            phases=[
                PhaseMeasurements(
                    voltage=load_L1_voltage,
                    power=load_L1_power
                ),
                PhaseMeasurements(
                    voltage=load_L2_voltage,
                    power=load_L2_power
                ),
                PhaseMeasurements(
                    voltage=load_L3_voltage,
                    power=load_L3_power
                )
            ],
            frequency=load_frequency
        )

        total_measurements = EnergyMeasurements(
            grid_bought=total_grid_bought,
            grid_sold=total_grid_sold,
            load_consumption=total_consumption,
            production=total_production,
            battery_charged=total_battery_charge,
            battery_discharged=total_battery_discharge
        )

        daily_measurements = EnergyMeasurements(
            grid_bought=daily_grid_bought,
            grid_sold=daily_grid_sold,
            load_consumption=daily_consumption,
            production=daily_production,
            battery_charged=daily_battery_charge,
            battery_discharged=daily_battery_discharge
        )

        pv_measurements = [
            PVStringMeasurements(
                voltage=pv1_voltage,
                current=pv1_current,
                power=pv1_power
            ),
            PVStringMeasurements(
                voltage=pv2_voltage,
                current=pv2_current,
                power=pv2_power
            )
        ]

        print("\ntotal")
        print(total_measurements)
        print("\ndaily")
        print(daily_measurements)
        print("\ngrid")
        print(grid_measurements)

        print(f" l1 power {grid_measurements.phases[0].power} l2 power {grid_measurements.phases[1].power} l3 power {grid_measurements.phases[2].power}")
        
        print("\nload")
        print(load_measurements)
        print("\nbattery")
        print(battery_measurements)
        print("\npv")
        print(pv_measurements)

        return InverterMeasurements(
            total=total_measurements,
            daily=daily_measurements,
            grid=grid_measurements,
            load=load_measurements,
            battery=battery_measurements,
            pv_strings=pv_measurements
        )


from pymodbus.client import AsyncModbusTcpClient
from pydeye.helper import ModbusMapper, BasicInfo, InverterType

class ModbusTCP():
    def __init__(self, host, port=502, unit=1):
        self.client = AsyncModbusTcpClient(host=host, port=port, timeout=1)
        self.unit = unit

    async def connect(self):
        await self.client.connect()

    async def close(self):
        self.client.close()

    def connected(self):
        return self.client.connected

    async def read_registers(self, register_address, count):
        if not self.connected():
            await self.connect()

        print(self.connected())

        print(f"Reading {count} registers at address {register_address}")
        result = await self.client.read_holding_registers(register_address, count=count, slave=self.unit)
        return result.registers

    async def write_registers(self, register_address, values):
        if not self.connected():
            await self.connect()

        result = await self.client.write_registers(register_address, values=values, slave=self.unit)
        return result
    
    async def get_basic_info(self) -> BasicInfo:
        data = await self.read_registers(0, 24)
        mapper = ModbusMapper(data, 0)
        mapper.dump()

        protocol_version = f"{mapper.get_value(2):04X}"
        serial_number = "".join([mapper.get_string(register) for register in range(3, 8)])
        type = InverterType.UNDEFINED
        for inverter_type in InverterType:
            if inverter_type.value[0] == mapper.get_value(0):
                type = inverter_type
                break

        rated_power = mapper.get_uint32(20)/10
        main_version = f"{mapper.get_value(14):04X}-{mapper.get_value(15):04X}-{mapper.get_value(11):04X}"
        hmi_version= f"{mapper.get_value(17):04X}-{mapper.get_value(18):04X}"

        #data = await self.read_registers(10056, 1)
        #mapper = ModbusMapper(data, 10056)
        # mapper.dump()

        return BasicInfo(type, serial_number, main_version, hmi_version, protocol_version, rated_power)
        
        
    

    


    


# Description: Custom exceptions for pyiskra


class NotAuthorised(Exception):
    pass


class DeviceNotSupported(Exception):
    pass


class DeviceConnectionError(Exception):
    pass


class InvalidResponseCode(Exception):
    pass


class DeviceNotSupported(Exception):
    pass


import re
from ..interfaces import ModbusTCP
from ..measurements import InverterMeasurements, EnergyMeasurements, BatteryMeasurements, PhaseMeasurements, ACMeasurements, PVStringMeasurements, Measurement
from ..helper import ModbusMapper
from ..exceptions import DeviceNotSupported


class BaseInverter:

    type: str
    serial_number: str
    main_version: str
    hmi_version: str
    protocol_version: str
    rated_power: int
    firmware: str
    adapter: ModbusTCP
    measurements: InverterMeasurements
    DEVICE_PARAMETERS = {}

    @staticmethod
    async def create_device(adapter):


        from .deye_SUN12KEU import DeyeSUN12KEU        
        return DeyeSUN12KEU(adapter)

        raise DeviceNotSupported(f"Unsupported device model: {type}")

    def __init__(self, adapter: ModbusTCP):
        self.adapter = adapter

    async def init(self):
        basic_info = await self.adapter.get_basic_info()
        self.type = basic_info.type
        self.serial_number = basic_info.serial_number
        self.main_version = basic_info.main_version
        self.hmi_version = basic_info.hmi_version
        self.protocol_version = basic_info.protocol_version
        self.rated_power = basic_info.rated_power

        print(f"Inverter {self.type} {self.rated_power}W {self.serial_number} initialized")


    async def update_status(self):
        """
        Updates the status of the device.

        This method needs to be re-defined in all sub-classes.
        """
        raise NotImplementedError


from dataclasses import dataclass

@dataclass
class Measurement:
    value: float
    unit: str

@dataclass
class EnergyMeasurements:
    grid_bought: Measurement
    grid_sold: Measurement
    load_consumption: Measurement
    production: Measurement
    battery_charged: Measurement
    battery_discharged: Measurement

@dataclass
class BatteryMeasurements:
    temperature: Measurement
    voltage: Measurement
    current: Measurement
    power: Measurement
    soc: Measurement


@dataclass
class PhaseMeasurements:
    voltage: Measurement
    current: Measurement
    power: Measurement

    def __init__(self, voltage, power, current=None):
        self.voltage = voltage
        self.current = current
        self.power = power

@dataclass
class ACMeasurements:
    total_power: Measurement
    phases: list[PhaseMeasurements]
    frequency: Measurement

@dataclass
class PVStringMeasurements:
    voltage: Measurement
    current: Measurement
    power: Measurement

@dataclass
class InverterMeasurements:
    total: EnergyMeasurements
    daily: EnergyMeasurements
    grid: ACMeasurements
    load: ACMeasurements
    battery: BatteryMeasurements
    pv_strings: list[PVStringMeasurements]


#!/usr/bin/env python3

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pydeye",
    version="0.0.5",
    description="Python deye inverter interface",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jakob Salmič",
    author_email="salmic.jakob@gmail.com",
    maintainer=", ".join(("Jakob Salmič <salmic.jakob@gmail.com>",)),
    license="GPL",
    url="https://github.com/UnknownHero99/pydeye",
    python_requires=">=3.8",
    packages=find_packages(),
    keywords=["homeautomation", "deye", "inverter", "modbus", "solar"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Home Automation",
    ],
    install_requires=["pymodbus"],
)
